require "pp"
require "erb"

def sizeof(typ)
  case typ
  when "integer"
    return 4
  when "real"
    return 4
  when "real(8)", "double precision"
    return 8
  when "logical"
    return 4
  else
    raise "type #{typ} is not defined"
  end
end

def ftype2ctype(typ,array,pointer)
  if array
    return "struct farray*"
  end
  case typ
  when "integer"
    c_type = "int"
  when "real"
    c_type = "float"
  when "real(8)", "double precision"
    c_type = "double"
  when "logical"
    c_type = "int"
  when /^character/
    c_type = "char"
    pointer = true
  else
    raise "type #{typ} is not defined"
  end

  c_type += "*" if pointer
  return c_type
end

def ftype2natype(typ)
  case typ
  when "integer","logical"
    na_type = "NA_LINT"
  when "real"
    na_type = "NA_SFLOAT"
  when "real(8)", "double precision"
    na_type = "NA_DFLOAT"
  else
    raise "type #{typ} is not defined"
  end
end

def get_cargs(args,out_array=nil)
  lw = args.length
  if out_array
    lw += 1
  end
  arg = args.collect{|arg|ftype2ctype(arg[:val_type],arg[:array]&&arg[:shape].include?(nil),true)}
  if out_array
    arg = ["struct farray*"] + arg
  end
  return [arg.join(", "),lw]
end


def str2len(str,saves)
  str_org = str
  new_str = ""
  while str && str.length > 0
    if str[0,1] == ":"
      new_str += ":"
      str = str[1..-1]
    elsif str[0,1] == "-"
      new_str += "-"
      str = str[1..-1]
    elsif str[0,1] == "+"
      new_str += "+"
      str = str[1..-1]
    elsif /\A\*\*\d+(.*)\z/ =~ str
      new_str << ("*" << new_str)
      str = $1
    elsif str[0,1] == "*"
      new_str += "*"
      str = str[1..-1]
    elsif str[0,1] == "/"
      new_str += "/"
      str = str[1..-1]
    elsif /^(\d+)([^\d]*.*)$/ =~ str
      new_str += $1
      str = $2
    elsif /^size\(([^\)]+)\)(.*)$/ =~ str
      ary = $1
      str = $2
      ary = ary.split(",")
      if ary.length == 1
        new_str += "(NA_TOTAL(#{ary[0].downcase}))"
      elsif ary.length == 2
        new_str += "(NA_STRUCT(#{ary[0].downcase})->shape[#{ary[1].to_i}])"
      else
        raise "bug"
      end
    elsif /^([^\d\-\+\*\/\(\)\:]+)((\d|\-|\+\*\/\(\)\:)*.*)$/ =~ str
      name = $1
      flg = false
      saves.length.times{|i|
        save = saves[i]
        if Hash === save
          if save[:name].downcase == name.downcase
            save[:used] = true
            flg = true
            break
          end
        else
          raise "bug"
        end
      }
      if flg
        new_str += name
      else
        new_str += "NUM2INT(rb_funcall(self,rb_intern(\"#{name}\"),0))"
      end
      str = $2
    elsif str[0,1] == "("
      new_str += "("
      str = str[1..-1]
    elsif str[0,1] == ")"
      new_str += ")"
      str = str[1..-1]
    else
      raise "bug: #{str}"
    end
  end

  if /:/ =~ new_str
    ary = new_str.split(":")
    case ary.length
    when 2
      return "(#{ary[1]}) - (#{ary[0]}) + 1"
    else
      return nil
    end
  else
    return new_str
  end
end

def r2c_func_narray(typ,name,shape)
  ctype = ftype2ctype(typ,false,true)
  str =<<"EOF"
  if ( !IsNArray(#{name}) )
    rb_raise(rb_eArgError,"#{name} must be NArray object");
  if ( NA_RANK(#{name}) != #{shape.length} )
    rb_raise(rb_eArgError,"rank of #{name} must be #{shape.length}");
EOF
  shape.each_with_index{|l,i|
    if l
      str =<<"EOF"
  if ( NA_STRUCT(#{name})->shape[#{i}] != #{l} )
    rb_raise(rb_eArgError,"shape[#{i}] of #{name} must be #{l}");
EOF
    end
  }
  str =<<"EOF"
  #{name} = na_cast_object(#{name}, #{ftype2natype(typ)});
EOF
  if shape.include?(nil)
    str +=<<"EOF"
#{wrap_fortran_array(ctype,name,shape)}
  c#{name}->array = NA_PTR_TYPE(#{name}, void*);
EOF
  else
    str +=<<"EOF"
  c#{name} = NA_PTR_TYPE(#{name}, #{ctype});
EOF
  end
  return str
end

def r2c_func(typ,name,pointer=false)
  case typ
  when "integer"
    return "#{pointer ?"*":""}c#{name} = NUM2INT(#{name})"
  when "real"
    return "#{pointer ?"*":""}c#{name} = (float)NUM2DBL(#{name})"
  when "real(8)"
    return "#{pointer ?"*":""}c#{name} = NUM2DBL(#{name})"
  when "logical"
    return "#{pointer ?"*":""}c#{name} = (#{name}==Qtrue)"
  when /^character/
    return "c#{name} = RSTRING_PTR(#{name})"
  else
    raise "type #{typ} is not defined"
  end
end

Type_code_ifort = {"double"=>10,"float"=>9,"int"=>3}
Type_code_gfortran = {"double"=>1, "float"=>1, "int"=>0}
def wrap_fortran_array(ctype,name,shape)
  ctype = ctype.sub(/\*$/,"")
  rank = shape.length
  str=<<"EOF"
  c#{name} = (struct farray*) xmalloc(sizeof(struct farray)+sizeof(struct fshape)*#{rank});
EOF
  str+=<<"EOF"
#if defined(IFORT) || defined(GFORTRAN)
{
  int i,j=1,offset=1;
  for (i=0;i<#{rank-1};i++) {
    j = j*NA_STRUCT(#{name})->shape[i];
    offset += j;
  }
  c#{name}->offset = -offset;
}
#endif
#if defined IFORT
  c#{name}->vsize = sizeof(#{ctype});
{
  long type = #{Type_code_ifort[ctype]};
  c#{name}->type = ((type<<(sizeof(void*)*6)) | 1);
}
  c#{name}->rank = #{rank};
#elif defined GFORTRAN
{
  c#{name}->flag = (#{rank} | (#{Type_code_gfortran[ctype]}<<4) | (sizeof(#{ctype})<<6));
}
#elif defined(FFC32) || defined(FFC64)
  c#{name}->total = NA_STRUCT(#{name})->total;
  c#{name}->total_byte = c#{name}->total*sizeof(#{ctype});
#endif
{
  int step=1, len, i;
  struct fshape* fs;
  for (i=0;i<#{rank};i++) {
    len = NA_STRUCT(#{name})->shape[i];
    fs = (struct fshape*)(c#{name}+1)+i;
#if defined(IFORT) || defined(FFC32) || defined(FFC64)
    fs->step = 8*step;
    fs->length = len;
    fs->flag = 1;
#endif
#if defined(FFC32) || defined(FFC64)
    fs->length2 = len;
#endif
#ifdef GFORTRAN
    fs->step = step;
    fs->start = 1;
    fs->end = len;
#endif
    step *= len;
  }
}
EOF
  return str
end

def c2r_func_narray(typ,name,shape,before,struct=false)
  str =<<"EOF"
{
  int shape[#{shape.length}];
EOF
  shape.each_with_index{|a,i|
    str +=<<"EOF"
  shape[#{i}] = #{a};
EOF
  }
  ctype = ftype2ctype(typ,false,true)
  str +=<<"EOF"
  #{name} = na_make_object(#{ftype2natype(typ)}, #{shape.length}, shape, cNArray);
EOF
  if before
    if struct || shape.include?(nil)
      str +=<<"EOF"
#{wrap_fortran_array(ctype,name,shape)}
  c#{name}->array = NA_PTR_TYPE(#{name}, void*);
EOF
    else
      str +=<<"EOF"
  c#{name} = NA_PTR_TYPE(#{name},#{ctype});
EOF
    end
  else
    str +=<<"EOF"
  #{ctype} cn#{name} = NA_PTR_TYPE(#{name},#{ctype});
  int i;
  for (i=0;i<NA_TOTAL(#{name});i++)
    cn#{name}[i] = c#{name}[i];
EOF
  end
  str += "}"
  str
end

def c2r_func(typ,name)
  case typ
  when "integer"
    return "#{name} = INT2NUM(c#{name})"
  when "real"
    return "#{name} = rb_float_new((double)c#{name})"
  when "real(8)"
    return "#{name} = rb_float_new(c#{name})"
  when "logical"
    return "#{name} = c#{name} ? Qtrue : Qnil"
  else
    raise "type #{typ} is not defined in #{name}"
  end
end

def dim2shape(dim,saves)
  shape = Array.new
  fl = 0
  str = ""
  while dim && dim.length > 0
    if /^(size\()(.*)$/ =~ dim
      fl = 1
      str += $1
      dim = $2
      next
    end
    c = dim[0,1]
    if fl > 0
      if c == "("
        fl += 1
      elsif c == ")"
        fl -= 1
      end
      str += c
    else
      if c == ","
        shape.push str2len(str,saves)
        str = ""
      else
        str += c
      end
    end
    dim = dim[1..-1]
  end
  shape.push str2len(str,saves)
  return shape
end

def parse_vars(str)
  ary = Array.new
  fl = 0
  name = ""
  dim = ""
  str.length.times{|i|
    c = str[i,1]
    if fl > 0
      if c == ")"
        fl -= 1
        if fl == 0
          next
        end
      elsif c == "("
        fl += 1
      end
      dim += c
      next
    else
      if c == "("
        fl += 1
        if fl == 1
          next
        else
          dim += c
          next
        end
      elsif c == ","
        name = name.strip.downcase
        if /^(\w+)=/ =~ name
          name = $1
        end
        ary.push [name, dim != "", dim]
        name = ""
        dim = ""
        next
      else
        name += c
        next
      end
    end
  }
  name = name.strip.downcase
  init = nil
  if /^(\w+)=(.*)$/ =~ name
    name = $1
    init = $2
  end
  ary.push [name, dim != "", dim, init]
  return ary
end

def parse_var_def(typ, str)
  if /^(.*),\s*dimension\((.+)$/ =~ typ
    typ = $1
    dim = $2
    if dim[-1,1] == ")"
      shape = dim[0..-2]
    else
      raise "bug: #{dim}"
    end
    ary = str.split(",").collect{|s|
      if /(.*)=>(.*)/ =~ str
        s = $1
      end
      [s.strip.downcase, true, shape, nil]
    }
  else
    ary = parse_vars(str)
  end
  return ary,typ
end







def parse_f90(fname)
  mod_name = nil
  includes = Array.new
  names = Array.new
  saves = Array.new
  func_saves = Hash.new
  objs = Array.new

  pre_line = nil
  contains = false
  lines = Array.new
  File.foreach(fname){|line|
    line = line.strip
    next if /^!/ =~ line
    if /^([^!]*)!/ =~ line
      line = $1
    end
    if pre_line
      line = pre_line + line
      pre_line = nil
    end
    if /^([^&]*)&/ =~ line
      pre_line = $1
      next
    end
    line = line.strip

    if /public(\s|:)*([^:]*)$/ =~ line
      $2.split(",").each{|str|
        str.strip!
        str = str[0,1].downcase + str[1..-1]
        names.push(str)
#        names.push(str) unless /ChangeResolution/ =~ str
        saves.delete(str)
      }
      next
    elsif /contains/ =~ line
      contains = true
      next
    elsif /save(\s|:)*(.*)$/ =~ line
      unless contains
        $2.split(",").each{|s|
          s = s.strip
          unless names.include?(s)
            saves.push s
          end
        }
      end
      next
    end

    lines += line.split(";").collect{|l| l.strip}
  }

  saves = saves.collect{|na| {:name => na} }

  func = nil
  vars = Array.new
  static_byte = 0
  dynamic_byte = 0
  interface = Array.new
  lines.each{|line|
    if func

      if /^end\s*#{func[:type].to_s}/ =~ line
        unless func[:skip]
          print "..end\n" if $DEBUG
          if func[:type] == :function || func[:type] == :subroutine
            func[:args].each{|arg|
              unless Hash === arg
                raise "cannot find definition of argment (#{arg}) of #{func[:name]}"
              end
            }
          end
          if interface.include? func[:name].downcase
            func[:interface] = true
          end
          objs.push func
        end
        func = nil
        next
      end

      next if func[:skip]

      if func[:type] == :interface
        if /module\s*procedure\s*(\w.*)$/ =~ line
          str = $1
          str.split(",").each{|name|
            name.strip!
            name = name[0,1].downcase + name[1..-1]
            names.push name
            interface.push name.downcase
          }
        end
        next
      end

      if /^(.*)::\s*(.*)$/ =~ line
        typ = $1.strip
        str = $2.strip
        inout = nil
        next if /allocatable/ =~ typ
        if /^(.+),\s*intent\(([^\)]+)\)(.*)$/ =~ typ
          typ = $1
          typ += $3 if $3
          inout = $2
          if func[:type] == :function && /out/ =~ inout
            raise "function must not have out or inout arguments: #{name} of #{func[:name]}"
          end
        end
        if /^(.*),\s*optional(.*)$/ =~ typ
          typ = $1
          typ += $3 if $3
          opt = true
        else
          opt = false
        end
        ary,typ = parse_var_def(typ, str)
        args = func[:args]
        ary.each{|name,array,shape,init|
          if (i = args.index(name))
            unless inout
              case func[:type]
              when :function
                warn "no intent in definition of #{str} of #{func[:name]} (function)"
                inout = "in"
              when :subroutine
                warn "no intent in definition of #{str} of #{func[:name]} (subroutine)"
                inout = "out"
              else
                raise "bug"
              end
            end
            args[i] = {:name => name, :inout => inout, :val_type => typ, :optional => opt, :array => array}
            if array
              if mod_name == "eee_module"
                args[i][:shape] = ["kmax"]
              else
                args[i][:shape] = dim2shape(shape,saves)
              end
            end
          elsif func[:type] == :function && func[:name].downcase == name.downcase
            func[:val_type] = typ
            func[:array] = array
            if array
              func[:shape] = dim2shape(shape,saves)
            end
          elsif inout
            raise "cannot find arg name (#{name}) of #{func[:name]} in [#{args.collect{|a|a[:name]}.join(",")}]"
          end
        }
        next
      end

      if /^allocate\((.+)\)$/ =~ line
        str = $1
        if /%(.*)$/ =~ str
          str = $1
        end
        ary = parse_vars(str)
        ary.each{|name,array,shape|
          vars.each{|v|
            if v[:name].downcase == name.downcase
              vars.delete(v)
              unless v[:array]
                raise "bug: #{v[:name]}"
              end
              v[:shape] = dim2shape(shape,saves)
              objs.push v
              break
            end
          }
        }
        next
      end

      saves.each_with_index{|ha,i|
        na = ha[:name]
        str = nil
        suffix = nil
        if /^#{na}\s*=\s*([^>].*)$/ =~ line
          str = $1.strip
        elsif /^#{na}([(%][^\s]+)\s*=\s*([^>].*)$/ =~ line
          str = $2
          suffix = $1.sub(/\(/,"[").sub(/\)/,"]").sub(/%/,".")
        end
        if str
          from = ""
          while str && str.length > 0
            if /^([\d]+\.*)(.*)$/ =~ str
              from += $1
              str = $2
              next
            elsif /^([\(\)\,])(.*)$/ =~ str
              from += $1
              str = $2
              next
            elsif /^([\s\+\-\*\/]+)(.*)$/ =~ str
              from += $1
              str = $2
              next
            elsif /^(\w+)\(([^\)]+)\)(.*)$/ =~ str
              a = $1
              v = $2
              str = $3
              from += "#{a}[#{v.gsub!(/([^\d\+\-\*\/\.]+)/,'c\1')}-1]"
              next
            elsif /^%(\w+\d*)(.*)$/ =~ str
              from += ".#{$1}"
              str = $2
              next
            elsif /^(\w+\d*)(.*)$/ =~ str
              vn = $1
              flag = false
              saves.each{|s|
                if s[:name] == vn
                  flag = true
                  break
                end
              }
              if flag
                from += "#{vn}"
              else
                from += "c#{vn}"
              end
              str = $2
            elsif /\A\.(true|false)\.\z/ =~ str
              from += $1
              break
            else
              raise "bug: #{str}"
            end
          end
          ha[:funcs] = true
          func_saves[func[:name]] ||= Array.new
          func_saves[func[:name]].push( {:name => ha[:name], :from => from, :suffix => suffix} )
          break
        end
      }

    else

      if /use\s*(\w+)/ =~ line
        name = $1
        includes.push(name) if /_module/ =~ name
        next
      end

      if /procedure/ !~ line && /module\s*(\w*)/ =~ line
        mod_name = $1
        next
      end

      if /^(.*)\s*::\s*(.*)$/ =~ line
        typ = $1.strip
        str = $2
        if /^(.*),\s*allocatable(.*)$/ =~ typ || /^(.*),\s*pointer(.*)$/ =~ typ
          alloc = true
          typ = $1
          typ += $2 if $2
        else
          alloc = false
        end
        ary,typ = parse_var_def(typ, str)
        ary.each{|name,array,shape,init|
          saves.delete_if{|ha|
            if ha[:name].downcase == name.downcase
              if typ=="integer"
                unless array
                  ha[:init] = init
                  ha[:type] = "int"
                  false
                end
              elsif typ=="real(8)"
                unless array
                  ha[:init] = init
                  ha[:type] = "double"
                  false
                end
              elsif mod_name == "ee_module"
                if ha[:name] == "params"
                  ha[:type] = "struct ee_param"
                  ha[:code] = <<"EOF"
struct ee_param {
  int im, jm, km, lm;
};
EOF
                  ha[:array] = 10
                  false
                end
              elsif mod_name == "eee_module"
                if ha[:name] == "params"
                  ha[:type] = "struct eee_param"
                  ha[:code] = <<"EOF"
struct eee_param {
  int im, jm, km, lm, mm, nm;
};
EOF
                  ha[:array] = 10
                  false
                end
              end
            else
              false
            end
          }
          names.each{|na|
            if na.downcase == name.downcase
              if alloc
                hash = {:name => na, :type => :variable, :val_type => typ, :array => array, :dynamic_byte => dynamic_byte}
                hash[:shape] = dim2shape(shape,saves)
                dynamic_byte += (3+shape.length*4)*4
                vars.push hash
              else
                hash = {:name => na, :type => :variable, :val_type => typ, :array => array, :static_byte => static_byte}
                if array
                  shape = dim2shape(shape,saves)
                  hash[:shape] = shape
                  len = 1
                  shape.each{|l|
                    unless l.to_i.to_s == l
                      raise "bug: #{l}"
                    end
                    len *= l.to_i
                  }
                  static_byte += sizeof(typ)*len
                else
                end
                objs.push hash
              end
              names.delete(na)
              break
            end
          }
        }
      end

      if /subroutine\s*([^\(]*)\(([^\)]*)\)/ =~ line
        name = $1
        flg = false
        names.each{|na|
          if na.downcase == name.downcase
            print "  subroutine: #{name}.." if $DEBUG
            args = $2.split(",").collect{|s| s.strip.downcase}
            func = {:name => na, :type => :subroutine, :args => args}
            names.delete(na)
            flg = true
            break
          end
        }
        unless flg
          (print "  skip subroutine: #{name}\n" unless func)  if $DEBUG
          func = {:skip => true, :type => :subroutine}
        end
        next
      end
      if /function\s*([^\(]*)\(([^\)]*)\)/ =~ line
        name = $1
        args = $2
        flg = false
        names.each{|na|
          if na.downcase == name.downcase
            print "  function: #{name}.." if $DEBUG
            args = args.split(",").collect{|s| s.strip.downcase}
            func = {:name => name, :type => :function, :args => args}
            names.delete(na)
            flg = true
            break
          end
        }
        unless flg
          (print "  skip function: #{name}\n" unless func) if $DEBUG
          func = {:skip => true, :type => :function}
        end
        next
      end
      if /interface\s*(\w+)/ =~ line
        name = $1
        flg = false
        names.each{|na|
          if na.downcase == name.downcase
            print "  interface: #{name}.." if $DEBUG
            func = {:name => name, :type => :interface}
            names.delete(na)
            flg = true
            break
          end
        }
        unless flg
          (print "  skip interface: #{name}\n" unless func) if $DEBUG
          func = {:skip => true, :type => :interface}
        end
        next
      end

    end

  }

  saves.delete_if{|save|
    if save[:used]
      if save[:funcs]
        false
      else
        raise "cannot find subscription for save #{save[:name]}"
      end
    elsif /eee?_module/ =~ mod_name
      case save[:name]
      when "nparams"
        save[:type] = "int"
        save[:init] = "-1"
        false
      when /params/
        func_saves.each{|fn,fs|
          fs.delete_if{|fs|
            /\Aparams\z/ =~ fs[:name] && (/\.(im|jm|km|lm|mm|nm)\z/ !~ fs[:suffix])
          }
        }
        false
      else
        func_saves.each{|fn,fs|
          fs.delete_if{|fs| fs[:name] == save[:name]}
        }
        true
      end
    else
      func_saves.each{|fn,fs|
        fs.delete_if{|fs| fs[:name] == save[:name]}
      }
      true
    end
  }


  if mod_name == "eee_module"
    objs.each do |obj|
      if obj[:name] == "eSpectralFromZeta"
        obj[:args].push( {:inout=>"in", :name => "kmax", :val_type => "integer"} )
      end
    end
  end

  return [mod_name, objs, includes, names, saves, func_saves]
end

def create_mod_ruby(mod_name, rmod_name, includes, wrapper_name)
  if /(\w*)_module/ =~ mod_name
    mod_name = $1
  end
  code =<<"EOF"
require 'numru/spml'
EOF
  if (includes.length + wrapper_name.length) > 0
    includes.each{|name|
      code +=<<"EOF"
require 'numru/spml/#{name}'
EOF
    }
    code +=<<"EOF"
module NumRu
  module SPML
    module #{rmod_name}
EOF
    includes.each{|name|
      if /(\w*)_module/ =~ name
        name = $1
      end
      code +=<<"EOF"
      extend #{name.upcase}
EOF
    }
    code +=<<"EOF"
    end
  end
end
EOF
  end

  return code
end


def create_mod_c(mod_name, rmod_name, objs, saves, func_saves)
  ifort_def = Array.new
  gfortran_def = Array.new
  ffc_def = Array.new
  ffc_asm32 = Array.new
  ffc_asm64 = Array.new
  main_code = Array.new
  init_code = Array.new

  objs.each{|obj|
    name = obj[:name]
    typ = obj[:type]
    args = obj[:args]
    in_args = Array.new
    in_args_opt = Array.new
    inout_args = Array.new
    out_args = Array.new
    out_in_args = Array.new
    case typ
    when :subroutine
      ret_type = "void"
      cargs,carg_lword = get_cargs(args)
      cargs = "(#{cargs})"
      args.each{|arg|
        if /in/ =~ arg[:inout].downcase
          if arg[:optional]
            in_args_opt.push arg
          else
            in_args.push arg
          end
        end
        if arg[:inout].downcase == "out"
          out_args.push arg
        elsif arg[:inout].downcase == "inout"
          inout_args.push arg
        end
      }
    when :function
      ret_type = ftype2ctype(obj[:val_type],obj[:array],false)
      cargs,carg_lword = get_cargs(args,obj[:array])
      cargs = "(#{cargs})"
      args.each{|arg|
        if /out/ =~ arg[:inout]
          raise "invalid inout"
        end
        if arg[:optional]
          in_args_opt.push arg
        else
          in_args.push arg
        end
      }
      hash = {:val_type => obj[:val_type], :name => "out", :array => obj[:array]}
      if obj[:array]
        hash[:shape] = obj[:shape]
        out_in_args.push hash
      else
        out_args.push hash
      end
    when :variable
      ret_type = ftype2ctype(obj[:val_type],false,obj[:array])
      cargs = ""
      hash = {:val_type => obj[:val_type], :name => "out", :array => obj[:array]}
      hash[:shape] = obj[:shape] if obj[:array]
      out_args.push hash
    when :interface
      next
    else
      raise "type is invalid"
    end
    ifort_name = "#{mod_name}_mp_#{name}_".downcase
    ifort_def.push "extern #{ret_type} #{ifort_name}#{cargs};"
    gfortran_name = "__#{mod_name}_MOD_#{name.downcase}"
    gfortran_def.push "extern #{ret_type} #{gfortran_name}#{cargs};"
#    ffc_name = "#{mod_name}.#{name}_"
    ffc_name = "#{mod_name}_#{name}_".downcase
    if typ == :variable
      ffc_asm32.push ERB.new(<<EOF,0,2).result(binding)
asm ("<%=ffc_name%>:");
asm ("  pushl %ebp");
asm ("  movl %esp, %ebp");
<%      if sb = obj[:static_byte] %>
<%        if obj[:array] %>
asm ("  leal MODULE.__<%=mod_name%>_+<%=sb%>, %eax");
<%        else %>
<%          case obj[:val_type] %>
<%          when "integer", "real","logical" %>
asm ("  movl MODULE.<%=mod_name%>_+<%=sb%>, %eax");
<%          when "real(8)" %>
asm ("  fildl MODULE.__<%=mod_name%>_+<%=sb%>");
<%          else %>
<%            raise "bug: #{obj[:val_type]}" %>
<%          end %>
<%        end %>
<%      elsif db = obj[:dynamic_byte] %>
asm ("  movl MODULE.<%=mod_name%>_+<%=db%>, %eax");
<%      end %>
asm ("  leave");
asm ("  ret");
EOF
      ffc_asm64.push ERB.new(<<EOF,0,2).result(binding)
asm ("<%=ffc_name%>:");
asm ("  push %rbp");
asm ("  movq %rsp, %rbp");
<%      if sb = obj[:static_byte] %>
<%        if obj[:array] %>
asm ("  leaq MODULE.<%=mod_name%>_@GOTPCREL(%rip), %r10");
asm ("  leaq <%=sb%>(%r10), %rax");
<%        else %>
asm ("  leaq MODULE.<%=mod_name%>_@GOTPCREL(%rip), %r10");
asm ("  leaq <%=sb%>(%r10), %rax");
<%        end %>
<%      elsif db = obj[:dynamic_byte] %>
asm ("  leaq MODULE.<%=mod_name%>_@GOTPCREL(%rip), %rax");
asm ("  movq <%=db*2%>(%rax), %r10");
asm ("  movq (%r10), %rax");
<%      end %>
asm ("  leave");
asm ("  ret");
EOF
      ffc_def.push "extern #{ret_type} #{ffc_name}();"
    else  # if typ
      unless obj[:interface]
        ffc_asm32.push ERB.new(<<EOF,0,2).result(binding)
asm ("<%=ffc_name%>:");
asm ("  pushl %ebp");
asm ("  movl %esp,%ebp");
<%        carg_lword.times{|l| %>
asm ("  movl <%=(carg_lword-l+1)*4%>(%ebp), %eax");
asm ("  pushl %eax");
<%        } %>
asm ("  call <%=mod_name%>.<%=name.downcase%>_");
asm ("  leave");
asm ("  ret");
EOF
        ffc_asm64.push ERB.new(<<EOF,0,2).result(binding)
asm ("<%=ffc_name%>:");
asm ("  push %rbp");
asm ("  movq %rsp,%rbp");
<%        nstack = [2**(Math::log((carg_lword > 6 ? carg_lword - 6 : 0)*8 + 16)/Math::log(2)).ceil, 16].max %>
asm ("  subq $<%=nstack%>,%rsp");
<%        if carg_lword > 6 %>
asm ("  movq %rbx,-8(%rbp)");
<%          (carg_lword - 6).times{|l| %>
asm ("  movq <%=(l+2)*8%>(%rbp),%r10");
asm ("  movq %r10,<%=l*8%>(%rsp)");
<%          } %>
<%        end %>
asm ("  call <%=mod_name%>.<%=name.downcase%>_@PLT");
<%        if carg_lword > 6 %>
asm ("  movq -8(%rbp),%rbx");
<%        end %>
asm ("  xorq %rax,%rax");
asm ("  addq $<%=nstack%>,%rsp");
asm ("  leave");
asm ("  ret");
EOF
        ffc_def.push "extern #{ret_type} #{ffc_name}#{cargs};"
      end
    end

    str = ERB.new(<<EOF,0,2).result(binding)
static VALUE
rbspml_<%=name%>(int argc, VALUE *argv, VALUE self)
{
<%    if ( in_args.length + in_args_opt.length + out_args.length + out_in_args.length ) > 0 %>
  VALUE <%=(in_args+in_args_opt+out_args+out_in_args).collect{|arg|arg[:name]}.join(", ")%>;
<%      (in_args + in_args_opt).each{|arg| %>
  <%=ftype2ctype(arg[:val_type],arg[:array]&&arg[:shape].include?(nil),arg[:array])%> c<%=arg[:name]%>;
<%      } %>
<%      out_in_args.each{|arg| %>
  <%=ftype2ctype(arg[:val_type],arg[:array],false)%> c<%=arg[:name]%>;
<%      } %>
<%      out_args.each{|arg| %>
<%        if arg[:name] == "out" %>
  <%=ftype2ctype(arg[:val_type],false,arg[:array])%> c<%=arg[:name]%>;
<%        else %>
  <%=ftype2ctype(arg[:val_type],arg[:array]&&arg[:shape].include?(nil),arg[:array])%> c<%=arg[:name]%>;
<%        end %>
<%      } %>
<%    end %>
EOF
    str += ERB.new(<<EOF,0,2).result(binding)

  rb_scan_args(argc, argv, <%=(['"'+in_args.length.to_s+in_args_opt.length.to_s+'"']+(in_args+in_args_opt).collect{|arg|"&"+arg[:name]}).join(", ")%>);
<%    in_args.each{|arg| %>
<%      if arg[:array] %>
<%=r2c_func_narray(arg[:val_type],arg[:name],arg[:shape])%>
<%      else %>
  <%=r2c_func(arg[:val_type],arg[:name])%>;
<%      end %>
<%    } %>
<%    in_args_opt.each{|arg| %>
  if ( <%=arg[:name]%> != Qnil )
<%      if arg[:array] %>
  <%=r2c_func_narray(arg[:val_type],arg[:name],arg[:shape])%>
<%      else %>
    <%=r2c_func(arg[:val_type],arg[:name],false)%>;
<%      end %>
<%    } %>
EOF

    if func_saves[obj[:name]]
      func_saves[obj[:name]].each{|ha|
        str += "  #{ha[:name]}"
        str += ha[:suffix] if ha[:suffix]
        str += " = #{ha[:from]};\n"
      }
    end

    out_in_args.each{|arg|
      if arg[:array]
        str +=<<"EOF"
#{c2r_func_narray(arg[:val_type],arg[:name],arg[:shape],true,true)};
EOF
      end
    }
    out_args.each{|arg|
      if arg[:array] && arg[:name] != "out"
        str +=<<"EOF"
#{c2r_func_narray(arg[:val_type],arg[:name],arg[:shape],true)};
EOF
      end
    }

    case typ
    when :subroutine, :function
      arg = args.collect{|a|
        an = a[:name]
        if a[:array] || /char/ =~ a[:val_type] 
          ca = "c#{an}"
        else
          ca = "&c#{an}"
        end
        if a[:optional] && a[:inout] != "out"
          ca = "#{an}==Qnil ? NULL : #{ca}"
        end
        ca
      }
      out = ""
      if typ == :function
        if obj[:array]
          arg = ['cout'] + arg
        else
          out = "cout = "
        end
      end
      arg = "(#{arg.join(", ")})"
    when :variable
      arg = ""
      if obj[:array]
        out = "cout = (#{ftype2ctype(obj[:val_type],false,true)})"
      else
        out = "cout = "
      end
    end

    str += ERB.new(<<EOF,0,2).result(binding)

#if defined IFORT
  <%=out%><%=ifort_name%><%=arg%>;
#elif defined GFORTRAN
  <%=out%><%=gfortran_name%><%=arg%>;
#elif defined(FFC32) || defined(FFC64)
<%    unless obj[:interface] %>
<%      arg = "()" if typ == :variable %>
  <%=out%><%=ffc_name%><%=arg%>;
<%    else %>
  rb_raise(rb_eRuntimeError, "<%=name%> is not supported for FFC comilar");
<%    end %>
#endif

EOF
    in_args.each{|arg|
      if arg[:array] && arg[:shape].include?(nil)
        str +=<<"EOF"
  free(c#{arg[:name]});
EOF
      end
    }
    in_args_opt.each{|arg|
      if arg[:array]
        str +=<<"EOF"
  if ( #{arg[:name]} != Qnil )
    free(c#{arg[:name]});
EOF
      end
    }
    out_in_args.each{|arg|
      if arg[:array]
        aname = arg[:name]
        ctype =ftype2ctype(arg[:val_type],false,true)
        str +=<<"EOF"
#if defined(FFC32) || defined(FFC64)
{
  #{ctype} cn#{aname} = NA_PTR_TYPE(#{aname},#{ctype});
  int i;
  for (i=0;i<NA_TOTAL(#{aname});i++)
    cn#{aname}[i] = ((#{ctype})c#{aname}->array)[i];
}
#endif
  free(c#{aname});
EOF
      end
    }

    str += ERB.new(<<EOF,0,2).result(binding)
<%    case typ %>
<%    when :function %>
<%      unless obj[:array] %>
  <%=c2r_func(obj[:val_type],'out')%>;
<%      end %>
  return out;
<%    when :subroutine %>
<%      out_args.each{|arg| %>
<%        unless arg[:array] %>
  <%=c2r_func(arg[:val_type],arg[:name])%>;
<%        end %>
<%      } %>
<%      ary = out_args + inout_args %>
<%      if ary.length == 1 %>
  return <%=ary[0][:name]%>;
<%      elsif ary.length > 0 %>
  return rb_ary_new3(<%=ary.length%>, <%=ary.collect{|arg|arg[:name]}.join(", ")%>);
<%      else %>
  return Qnil;
<%      end %>
<%    when :variable %>
<%      if obj[:array] %>
<%=c2r_func_narray(obj[:val_type],'out',obj[:shape],false)%>;
<%      else %>
  <%=c2r_func(obj[:val_type],'out')%>;
<%      end %>
  return out;
<%    else %>
<%      raise "invalid type: #{typ}" %>
<%    end %>
}
EOF
    main_code.push str

    if /^#{rmod_name.downcase}_Initial$/ =~ name
      mname = "initial"
    else
      mname = name
    end
    init_code.push "  rb_define_module_function(m#{rmod_name}, \"#{mname}\", rbspml_#{name}, -1);"
    init_code.push "  rb_define_method(m#{rmod_name}, \"#{mname}\", rbspml_#{name}, -1);"
  }

  code =<<"EOF"
#include "ruby-spml.h"

EOF
  if ifort_def.length > 0
    code += "#ifdef IFORT\n"
    code += ifort_def.join("\n")
    code += "\n#endif\n"
  end
  if gfortran_def.length > 0
    code += "#ifdef GFORTRAN\n"
    code += gfortran_def.join("\n")
    code += "\n#endif\n"
  end
  if ffc_def.length > 0
    code += "#ifdef FFC32\n"
    code += ffc_asm32.join("")
    code += "#endif\n"
    code += "#ifdef FFC64\n"
    code += ffc_asm64.join("")
    code += "#endif\n"
    code += "#if defined(FFC32) || defined(FFC64)\n"
    code += ffc_def.join("\n")
    code += "\n#endif\n"
  end
  code += "\n\n"
  if saves.length > 0
    saves.each{|save|
      code += save[:code] if save[:code] 
      code += "static #{save[:type]} #{save[:name]}"
      code += "[#{save[:array]}]" if save[:array]
      code += " = #{save[:init]}" if save[:init]
      code += ";\n"
    }
    code += "\n\n"
  end
  if main_code.length > 0
    code += main_code.join("\n")
  end
  code += ERB.new(<<EOF,0,2).result(binding)


void init_<%=mod_name%>(VALUE mSPML)
{
  VALUE m<%=rmod_name%>;

  m<%=rmod_name.upcase%> = rb_define_module_under(mSPML, "<%=rmod_name%>");

<%  if init_code.length > 0 %>
<%=init_code.join("\n")%>
<%  end %>
}
EOF

  return code
end


def create_header
  code =<<"EOF"
#include <stdlib.h>
#include "ruby.h"
#include "narray.h"

struct fshape {
#if defined IFORT
  long length;
  long step;
  long flag;
#elif defined GFORTRAN
  long step;
  long start;
  long end;
#elif defined(FFC32) || defined(FFC64)
  long flag;
  long length;
  long step;
  long length2;
#endif
};
struct farray {
#ifdef IFORT
  void* array;
  long vsize;
  long offset;
  long type;
  long rank;
  long dummy;
#elif defined GFORTRAN
  void *array;
  long offset;
  long flag;
#elif defined(FFC32) || defined(FFC64)
  void* array;
  long total_byte;
  long total;
#endif
};
EOF
  return code
end

path = ARGV.shift
unless path
  raise "Usage: ruby #$0 path_to_spml_source_dir"
end


#srcdir = File.join(File.directory(__FILE__),"../ext")
srcdir = File.join(File.dirname(__FILE__),"..")

mod_names = Array.new
Dir[File.join(path,"src","*_module.f90")].each{|fname|

  print fname+"\n"

  mod_name, objs, includes, wrapper_name, saves, fsaves = parse_f90(fname)
  rmod_name = mod_name.sub(/_module/,"").upcase

  fname = "lib/spml/#{mod_name}.rb"
  File.open(fname,"w"){|file|
    file.print create_mod_ruby(mod_name, rmod_name, includes, wrapper_name)
  }
  fname = File.join(srcdir,"#{mod_name}.c")
  File.open(fname,"w"){|file|
    file.print create_mod_c(mod_name, rmod_name, objs, saves, fsaves)
  }

  mod_names.push mod_name
}

fname = File.join(srcdir,"ruby-spml.h")
File.open(fname,"w"){|file|
  file.print create_header
}

fname = "lib/spml.rb"
File.open(fname,"w"){|file|
  file.print <<"EOF"
require "narray"
require "numru/spml.so"
EOF
  mod_names.each{|name|
  file.print <<"EOF"
require "numru/spml/#{name}"
EOF
  }
}

fname = File.join(srcdir,"init.c")
File.open(fname,"w"){|file|
  file.print <<"EOF"
#include "ruby.h"
#include "narray.h"

EOF

  mod_names.each{|name|
    file.print "void init_#{name}(VALUE);\n"
  }

  file.print <<"EOF"
void Init_spml(){
  VALUE mNumRu, mSPML;

  mNumRu = rb_define_module("NumRu");
  mSPML = rb_define_module_under(mNumRu, "SPML");

EOF

  mod_names.each{|name|
    file.print "  init_#{name}(mSPML);\n"
  }

  file.print <<"EOF"
}
EOF

  file.print <<"EOF"
int abort_(){
  rb_raise(rb_eRuntimeError, "Error occured in the Fortran");
}
EOF

}
